/*---------------------------------------------------------------------------
 *
 *      Title:          Memory Allocation tracking and checking
 *
 *      File:           dyn_val.c
 *
 *      Creation Date:  August 1992
 *
 *      Description:
 *             Provide replacement for standard allocation routines which track
 *      all allocations, provide error notification if freeing non-allocated memory
 *      allow dumping of what memory is in use (good for checking for leaks), and
 *      automatically verify for overwrites past start & end.
 *
 *      Data sources:
 *
 *      1.0     R.J.Harrison    Implement
 *      1.1     R.J.Harrison    Modifications for speed          February 1993
 *      1.2     R.J.Harrison    Complete re-write                November 1994
 *      1.3     R.J.Harrison    Generic PC version               August   1995
 *      1.4     R.J.Harrison    Shareware release changes        June     1996
 *
 *      Copyright  1991-1996 R.J.Harrison
 *---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include "windows.h"
#endif

#ifdef calloc
#undef calloc
#endif
#ifdef free
#undef free
#endif
#ifdef malloc
#undef malloc
#endif
#ifdef savestr
#undef savestr
#endif

void DV_error_log (char *,...);

#define DO_CHECK
/*
 * #define CREATE_ALLOC_LOG_FILE
 */

static long check_key = 0Xfeaee1e1, check_key_long[] =
{0xaabbccdd, 0x11223344};

typedef struct CHECK
{
    long pre[2];
    void *data;
    long index;
    char *checkword;
}
CHECK;

typedef struct CHECK_SHORT
{
    void *data;
    long line, size;
    char *name;
}
CHECK_SHORT;

typedef enum
{
    C_ADD, C_REMOVE, C_SIZE, C_USAGE, C_USAGE_ALL, C_FIND, C_FIND_I, C_CHECK, C_DUMP, C_REMOVE_I
} CHECK_ACTION;

char *op_ascii(void *input, int op_size, int max_size)
{
	static char buff[40];
	char *op = buff, *pl = (char*)input;


	if (op_size >= max_size)
		op_size = max_size;

	if (op_size > (sizeof(buff)-2))
		op_size = sizeof(buff)-2;

	while (op_size--)
		if (isprint(*pl))
			*op++ = *pl++;
		else
			pl++;
	*op=0;
	return buff;

}

/*
 * Smarter version of realloc which will work with a null pointer 
 * signifying to create the list...
 */
void *sm_realloc (void *ptr, size_t ne, size_t sz)
{
    if (!ptr)
	return calloc (ne, sz);
    else
	return realloc (ptr, ne * sz);
}

static long check_memory_item (void *data, CHECK_SHORT * cs)
{
#ifdef _WIN32
	if (IsBadReadPtr(data, 4))
		return 0;
#endif
	if(data)
	{
	    char *td = data;
	    CHECK *ck;


	    td = td - sizeof (CHECK);
	    ck = (CHECK *) td;

	    if (memcmp (ck->checkword, &check_key, sizeof (check_key)))
	    {
	        FILE *ol = fopen("mem.ol","a+");
		DV_error_log ("memory %lx, %s/%ld overwrite past end.\n", cs->data, cs->name, cs->line);
	        if(ol)
		{
	            fprintf (ol,"memory %lx, %s/%ld overwrite past end.\n", cs->data, cs->name, cs->line);
	            fclose(ol);
		}
	        memcpy (ck->checkword, &check_key, sizeof (check_key));

		return (1);
	    }

	    if (memcmp (ck->pre, check_key_long, sizeof (ck->pre)))
	    {
	        FILE *ol = fopen("mem.ol","a+");
		DV_error_log ("memory %lx, %s/%ld overwrite at start.\n", cs->data, cs->name, cs->line);
	        if(ol)
		{
	            fprintf (ol,"memory %lx, %s/%ld overwrite at start.\n", cs->data, cs->name, cs->line);
	            fclose(ol);
		}
	        memcpy (ck->pre, check_key_long, sizeof (ck->pre));
		return (1);
	    }
	}
    return (0);
}

static long do_check_list (CHECK_ACTION action, void *ptr, CHECK_SHORT * cs)
{
    static CHECK_SHORT *check_list = 0;
    static long *first_free, list_size = 0, list_max_size = 0, first_free_pos = -1,
      list_inc = 700;

    switch (action)
    {
	    case C_ADD:
		{
		    long pos = 0;

		    if (first_free_pos >= 0)
			pos = first_free[first_free_pos--];

		    if (!pos)
		    {
			if (list_size >= list_max_size)
			{
			    list_max_size += list_inc;
			    check_list = sm_realloc (check_list, list_max_size, sizeof (CHECK_SHORT));
			    first_free = sm_realloc (first_free, list_max_size, sizeof (*first_free));
			}
			pos = list_size++;
		    }
		    if (!check_list)
			return (0);

		    check_list[pos] = *cs;
		    check_list[pos].data = ptr;
		    return (pos);

		    break;
		}

	    case C_REMOVE:
		ptr = (void *) do_check_list (C_FIND_I, ptr, 0);

	    case C_REMOVE_I:
		{
		    long remove_index = (long) ptr;

		    if ((remove_index < 0) || (remove_index > list_size))
			return -1;

		    if (remove_index > -1)
		    {
			first_free[++first_free_pos] = remove_index;
	/*
	 * you will not use memory which has been freed ...
	 */
			memset (check_list[remove_index].data, 0xfe, check_list[remove_index].size);
			memset (check_list[remove_index].data, 0, 1);
			memset (&check_list[remove_index], 0, sizeof (CHECK_SHORT));
		    }
		    break;
		}

	    case C_FIND_I:
		{
		    long i;

		    for (i = 0; i < list_size; i++)
			if (check_list[i].data == ptr)
			    return i;

		    return -1;
		    break;
		}

	    case C_FIND:
		{
		    long i = do_check_list (C_FIND_I, ptr, 0);

		    if (i < 0)
		    {
			memset (cs, 0, sizeof (*cs));
			cs->name = "Unknown";
	 	        return 0;
		    }
		    else
			*cs = check_list[i];
			return 1;
		    break;
		}

	    case C_SIZE:
		{
		    return list_size;
		    break;
		}

	    case C_USAGE:
		{
		    return (long) (list_size * (sizeof (CHECK_SHORT) + sizeof (long)));
		    break;
		}

	    case C_USAGE_ALL:
		{
		    long usage = 0, i;

		    for (i = 0; i < list_size; i++)
			usage += check_list[i].size;

	            return (usage);
		    break;
		}

	    case C_CHECK:
		{
		    long failed = 0, i;

		    for (i = 0; i < list_size; i++)
		    {
			failed += check_memory_item (check_list[i].data, &check_list[i]);
		    }
		    if (failed)
			DV_error_log  ("%ld memory corruptions found\n", failed);

		    return failed;
		    break;
		}

	    case C_DUMP:
		{
		    long i;
		    static int cn=1;
		    char fn[50];
		    FILE *op;
	    
		    sprintf (fn, "memory%02d.usage",cn++);
		    op = fopen (fn, "w");

		    if (!op)
			return 1;

		    fprintf (op, "Detailed dynamic memory usage\n");
		    for (i = 0; i < list_size; i++)
		    {
	                if(check_list[i].data)
	  	     	    fprintf (op, "%-20s :%4ld size %ld, [%s]\n", 
	                             check_list[i].name, check_list[i].line, check_list[i].size,
								 op_ascii(check_list[i].data,20,check_list[i].size));
		    }
		    fclose (op);
		    return 0;
		    break;
		}
    }
	return 0;
}

static CHECK *start;
static long memory_used = 0;

static char *get_checkptr (void *ptr, long sz)
{
    char *p = ptr;
    return (p + sz);
}

static char *get_short_name(char *module)
{
char *p;
            /*
             * Now remove pathnames from the module name
             */
            p = strrchr(module,'\\');

            if (!p)
                p = strrchr(module,'/');

            if (p)
                return ++p;
            else
                return module;
}
/*
 * 'Remember' a given chunk of memory.
 * ****NOTE*** the CHECK portion must have been zero'd already.
 */
static void DV_remeber (CHECK * cptr, char *module, long line, long sz)
{
    if (cptr)
    {
	void *ptr = (char *) cptr + sizeof (CHECK);
	char *cp = get_checkptr (ptr, sz);
	CHECK *ck = cptr;
	CHECK_SHORT cs;

	memcpy (cp, &check_key, sizeof (check_key));

	if (ck)
	{
#ifdef CREATE_ALLOC_LOG_FILE
	    static FILE *mem_log = 0;
	    if (!mem_log)
		mem_log = fopen ("mem.log", "w");
	    if (mem_log)
		fprintf (mem_log, "%8x: %8x %-12s %4d\n", ptr, ck, module, line);
#endif

	    memcpy (ck->pre, &check_key_long, sizeof (ck->pre));


	    cs.name = get_short_name(module);
	    cs.line = line;
	    cs.size = sz;

	    ck->data = ptr;
	    ck->checkword = cp;
	    ck->index = do_check_list (C_ADD, ptr, &cs);

#ifdef KEEP_LIST
	    if (!start)
		start = ck;
	    else
	    {
		ck->next = start;
		start->prev = ck;
		start = ck;
	    }
#endif
	}
    }
}


static long DV_remove (void *data, char *module, long line)
{
    CHECK *ck;
    char *td = data;

    td = td - sizeof (CHECK);
    ck = (CHECK *) td;

#ifdef _WIN32
	if (IsBadReadPtr(data, sizeof(CHECK)))
		return 0;
#endif

    if (!data || !ck->data)
	return (0);

    if (ck->data == data)
    {
		if (memcmp (ck->checkword, &check_key, sizeof (check_key)))
		{
		    CHECK_SHORT cs;
		    if (do_check_list (C_FIND, ck->data, &cs))
		    {
		    DV_error_log ("Memory check %s (line %ld) failed\noverwrite with 0x%x",
			       cs.name, cs.line, *ck->checkword);
		    }
		    else
				return 0;
		}

		if (do_check_list (C_REMOVE_I, (void *) ck->index, 0) < 0)
		    return (0);

		ck->data = 0;
		free (ck);
		return (1);
    }
    else
    {
		CHECK_SHORT cs;
		if (do_check_list (C_FIND, data, &cs))
		{
			DV_error_log ("Overwrite past start of memory block %s (line %ld)",
			   cs.name, cs.line);
			return 1; // Freeing a valid but corrupted block
		}
    }
    return (0);
}

static char *oph (char *ptr, char *opbuf, long sz)
{
    static char c[10];
    *opbuf = 0;
    while (sz--)
    {
	sprintf (c, "%02x", *ptr++);
	strcat (opbuf, c);
    }
    return (opbuf);
}


void DV_ov_check_memory (void)
{
    do_check_list (C_CHECK, 0, 0);
}

void DV_scan_check_memory (void)
{
    static int t = 0, in_progress=0;

    if (in_progress)
        return;

    in_progress = 1;

    if (!t || t <= time (0))
	do_check_list (C_CHECK, 0, 0);
    t = time (0) + 1;

    in_progress = 0;
}


long DV_get_dynamic_memory_usage (void)
{
#ifdef AMIGA
    extern long bss_size;
#else
    long bss_size = 62122;
#endif

#ifdef DO_CHECK
    long sz = do_check_list (C_USAGE, 0, 0);

    return do_check_list (C_USAGE_ALL, 0, 0) + bss_size;
#else
    return (memory_used + bss_size);
#endif
}

void DV_dump_memory (void)
{
    if (do_check_list (C_DUMP, 0, 0))
		DV_error_log ("Unable to open memory dump file");
}


void *DV_get (unsigned long sz)
{
    void *ptr;
    if (sz && sz > 0)
    {
	long size = sizeof (CHECK) + sz + sizeof (long);

	ptr = malloc (size);
	if (ptr)
	{
	    if (ptr)
		memset (ptr, 0, size);
	    ptr = (char *) ptr + sizeof (CHECK);
	    return (ptr);
	}
    }
    return (NULL);
}

void DV_track (unsigned long sz, void *ptr, char *module, long line)
{
    void *aptr;
    if (!ptr)
        printf ("failed new, %s, %d",module,line);

	if (ptr)
	{
	    aptr = ptr;
	    ptr = (char *) ptr - sizeof (CHECK);
	    DV_remeber (ptr, module, line, sz);
	}
}


/*
 * Alloc sanity checking routines
 */
void *DV_malloc (unsigned long sz, char *module, long line)
{
#ifdef DO_CHECK
    void *ptr, *aptr;
    if (sz && sz > 0)
    {
	long size = sizeof (CHECK) + sz + sizeof (long);

	ptr = malloc (size);
	if (ptr)
	{
	    if (ptr)
		memset (ptr, 0, size);
	    aptr = ptr;
	    ptr = (char *) ptr + sizeof (CHECK);
	    DV_remeber (aptr, module, line, sz);
	    return (ptr);
	}
    }
    return (NULL);
#else
    return (malloc (sz));
#endif
}

void *DV_calloc (unsigned long ne, unsigned long sz, char *module, long line)
{
#ifdef DO_CHECK
    void *ptr = DV_malloc (ne * sz, module, line);
    return (ptr);
#else
      return (calloc (ne, sz));
#endif
}

void *DV_savestr (char *p, char *module, long line)
{
    char *c = DV_malloc (strlen (p) + 3, module, line);
    strcpy (c, p);
    return (c);
}

void *DV_tempnam (char *dir, char *prefix, char *module, long line)
{
#ifdef DO_CHECK
    char *ptr, *ret_ptr;
    if (ptr = tempnam(dir,prefix))
    {
    ret_ptr = DV_savestr(ptr, module, line);
    free (ptr);
    return (ret_ptr);
    }
    else
        return (NULL);
#else
    return (tempnam(dir,prefix));
#endif
}

void *DV_tmpnam (char *str, char *module, long line)
{
#ifdef DO_CHECK
    char *ptr, *ret_ptr;
    if (ptr = tmpnam(str))
    {
    ret_ptr = DV_savestr(ptr, module, line);
    free (ptr);
    return (ret_ptr);
    }
    else
        return (NULL);
#else
    return (tempnam(dir,prefix));
#endif
}

void DV_free (void *a, char *module, long line)
{
#ifdef DO_CHECK
    if (!DV_remove (a, module, line))
	DV_error_log ("free(): Module %s (line %ld) failed\nattempt to free non-alloc memory %8x", 
	           get_short_name(module), line, a);
#else
    free (a);
#endif
}

void *DV_realloc (void *blk, unsigned long new_sz, char *module, long line)
{
#ifdef DO_CHECK
    void *ptr = DV_malloc (new_sz, module, line);

    if(blk)
    {
        CHECK_SHORT cs;
        do_check_list (C_FIND, blk, &cs);
        memcpy (ptr, blk, min(cs.size, new_sz));
        DV_free (blk, module, line);
    }
    return (ptr);
#else
      return (realloc (blk, new_sz));
#endif
}


